import heapq
import collections
import itertools
from operator import itemgetter

'''
>>> # Convenient definitions
>>> from cryptonita.fuzzy_set import *

'''

class FuzzySet(dict, collections.Set):
    # TODO talk about this!
    def __init__(self, iterable=(), pr=None, min_membership=0.0):
        ''' Create a FuzzySet from an iterable.

            If you have a dictionary where the values represent the
            probability or membership of the key you can use it directly

            >>> FuzzySet({'a': 1, 'b': 0.8, 'c': 0.4})
            {'a' -> 1.0000, 'b' -> 0.8000, 'c' -> 0.4000}

            You can cut the set setting a minimum for the membership:

            >>> FuzzySet({'a': 1, 'b': 0.8, 'c': 0.4}, min_membership=0.5)
            {'a' -> 1.0000, 'b' -> 0.8000}

            Note that the min_membership effect will persist.

            If you don't have a dict, you can use two iterables

            >>> FuzzySet(['a', 'b', 'c'], [1, 0.8, 0.4], min_membership=0.5)
            {'a' -> 1.0000, 'b' -> 0.8000}

            or you can use this python trick to unpack a list of tuples and
            get the same effect:

            >>> FuzzySet(*zip(*[('a', 1), ('b', 0.8), ('c', 0.4)]), min_membership=0.5)
            {'a' -> 1.0000, 'b' -> 0.8000}

            or, as a shortcut, set pr to 'tuple'.

            >>> FuzzySet([('a', 1), ('b', 0.8), ('c', 0.4)], pr='tuple', min_membership=0.5)
            {'a' -> 1.0000, 'b' -> 0.8000}

            You can build a fuzzy set from a set of keys. In this case you
            need to set a default probability or membership for all the keys:

            >>> FuzzySet(['a', 'b', 'c'], pr=0.8)
            {'c' -> 0.8000, 'b' -> 0.8000, 'a' -> 0.8000}

            If the default is omitted, 1.0 is used:

            >>> FuzzySet(['a', 'b', 'c'])
            {'c' -> 1.0000, 'b' -> 1.0000, 'a' -> 1.0000}

        '''

        self.min_membership = min_membership
        self._check_probability(None, min_membership, True)

        if not isinstance(iterable, dict):
            if pr is None:
                iterable = ((k, 1.0) for k in iterable)

            elif isinstance(pr, (float, int)):
                pr = float(pr)
                iterable = ((k, pr) for k in iterable)

            elif pr == 'tuple':
                iterable = iterable

            else:
                iterable = zip(iterable, pr)

            iterable = ((k, pr) for (k, pr) in iterable if pr > min_membership)

        else:
            iterable = ((k, pr) for (k, pr) in iterable.items() if pr > min_membership)

        dict.__init__(self, iterable)
        [self._check_probability(k, pr) for k, pr in self.items()]

    def __repr__(self):
        items = sorted(self.items(), key=itemgetter(1, 0), reverse=True)
        return '{' + ', '.join('%s -> %0.4f' % (repr(k),p) for k,p in items) + '}'

    def _check_probability(self, elem, prob, about_min_membership=False):
        if not (0 <= prob <= 1):
            if about_min_membership:
                raise ValueError("The minimum membership is %0.4f but it must be a value between 0 and 1."
                                        % (prob))
            else:
                raise ValueError("The membership of %s is %0.4f but it must be a value between 0 and 1."
                                        % (repr(elem), prob))

    def most_likely(self, n=None):
        '''
            >>> g = FuzzySet(['a', 'b', 'c'], [0.9, 0.8, 0.7])

            >>> g
            {'a' -> 0.9000, 'b' -> 0.8000, 'c' -> 0.7000}

            >>> g.most_likely()
            'a'

            >>> g.most_likely(2)
            ('a', 'b')

            '''
        nlargest = heapq.nlargest(1 if n == None else n, self.items(), key=itemgetter(1))
        if n == None:
            return nlargest[0][0]
        else:
            return tuple(zip(*nlargest))[0]

    def __setitem__(self, elem, prob):
        r'''Add an element to the set and assign it a value of <prob>.

            >>> g = FuzzySet()

            >>> g['a'] = 0.51
            >>> g['b'] = 1
            >>> g['c'] = 0.7

            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000, 'a' -> 0.5100}

            We support updates

            >>> g['a'] = 0.3
            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000, 'a' -> 0.3000}

            The value of <prob> is the probability of the element to be in the
            set. Therefore it must be a value between 0 and 1.

            Zero values are discarded:
            >>> g['d'] = 0
            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000, 'a' -> 0.3000}

            Greater than 1 (or negative) raise an exception:
            >>> g['e'] = 42
            Traceback (most recent call last):
            <...>
            ValueError: The membership of 'e' is 42.0000 but it must be a value between 0 and 1.

            The FuzzySet can be configured to drop and not save any key lower
            than a given threshold:

            >>> g.min_membership = 0.4
            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000}

            >>> g['d'] = 0.1
            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000}

            >>> g['c'] = 0.1
            >>> g
            {'b' -> 1.0000}

        '''

        self._check_probability(elem, prob)
        if prob < self.min_membership or prob == 0:
            if elem in self:
                del self[elem]

            # drop and don't save
            return

        dict.__setitem__(self, elem, prob)

    @property
    def min_membership(self):
        return self._min_membership

    @min_membership.setter
    def min_membership(self, pr):
        self._check_probability(None, pr, True)
        self._min_membership = pr

        self.cut_off(self._min_membership)

    def __getitem__(self, elem):
        return dict.get(self, elem, 0)

    def cut_off(self, n):
        r'''If <n> is an integer, drop all the elements except
            the <n> most likely ones.

            >>> g = FuzzySet(['a', 'b', 'c'], [0.9, 0.8, 0.7])

            >>> g
            {'a' -> 0.9000, 'b' -> 0.8000, 'c' -> 0.7000}

            >>> g.cut_off(2)
            >>> g
            {'a' -> 0.9000, 'b' -> 0.8000}

            If <n> is between 0 and 1, drop all the elements which
            probability is lower than <n>.

            >>> g.cut_off(0.85)
            >>> g
            {'a' -> 0.9000}

        '''

        if isinstance(n, int):
            nlargest = heapq.nlargest(n, self.items(), key=itemgetter(1))
            self.clear()
            dict.update(self, nlargest)

        else:
            for k in list(self.keys()):
                if self[k] < n:
                    del self[k]



    def update(self, other):
        r'''Perform the union of this set and the <other> and update self
            with the union.

            >>> g = FuzzySet(['a', 'b'], 0.5)

            >>> g.update(FuzzySet(['a', 'c'], [1, 0.2]))
            >>> g
            {'a' -> 1.0000, 'b' -> 0.5000, 'c' -> 0.2000}

        '''

        tmp = self.union(other)
        self.clear()
        dict.update(self, tmp)

    def intersection(self, other):
        r'''Return a fuzzy sets with elements that are present in both
            sets with the minimum of both membership.

            >>> a = FuzzySet(['a', 'b'], (0.7, 0.4))
            >>> b = FuzzySet(['a', 'b', 'c'], (1.0, 0.2, 0.1))

            >>> a.intersection(b)
            {'a' -> 0.7000, 'b' -> 0.2000}

            >>> a & b
            {'a' -> 0.7000, 'b' -> 0.2000}

        '''

        tmp = FuzzySet()
        min_set = other if len(other) < len(self) else self
        max_set = other if min_set is self else self

        for k, pr in min_set.items():
            tmp[k] = min(max_set[k], pr)

        return tmp

    def union(self, other):
        r'''Return a fuzzy sets with elements of both
            sets with the maximum of both membership.

            >>> a = FuzzySet(['a', 'b'], (0.7, 0.4))
            >>> b = FuzzySet(['a', 'b', 'c'], (1.0, 0.2, 0.1))

            >>> a.union(b)
            {'a' -> 1.0000, 'b' -> 0.4000, 'c' -> 0.1000}

            >>> a | b
            {'a' -> 1.0000, 'b' -> 0.4000, 'c' -> 0.1000}

        '''
        max_set = other if len(other) > len(self) else self
        min_set = other if max_set is self else self

        tmp = FuzzySet(max_set)
        for k, pr in min_set.items():
            tmp[k] = max(tmp[k], pr)

        return tmp

    def issubset(self, other):
        r'''Return True if self is a subset of the fuzzy set <other>.
            Be a subset means that all the elements in self
            are present in <other> and with a lesser membership
            value.

            >>> a = FuzzySet(['a', 'b'], (0.7, 0.4))
            >>> b = FuzzySet(['a', 'b', 'c'], (1.0, 0.2, 0.1))

            >>> a.issubset(b)
            False

            >>> a['b'] = 0.2
            >>> a.issubset(b)
            True

            >>> b.issuperset(a)
            True

            In other words, A is a subset of B if A union B yields B

            >>> a.union(b) == b
            True

        '''
        for k, pr in self.items():
            if pr > other[k]:
                return False

        return True

    def issuperset(self, other):
        return other.issubset(self)

    def __and__(self, other):
        return self.intersection(other)

    def __or__(self, other):
        return self.union(other)


def join_fuzzy_sets(iterable, cut_off, j):
    r'''
        Join the given fuzzy sets and return a single one.

        Combine all the elements of all the sets, in order, and
        build all the possible strings from that combination,
        joining the pieces with <j>.

        Each string it has a probability equals to the multiplication
        of the individual probabilities of each piece.

        For example:

        >>> A = FuzzySet(dict(y=0.9, a=0.6))
        >>> B = FuzzySet(dict(j=0.7, e=0.6))
        >>> C = FuzzySet(dict(s=0.4))

        >>> join_fuzzy_sets([A, B, C], cut_off=0.0, j='')
        {'yjs' -> 0.2520, 'yes' -> 0.2160, 'ajs' -> 0.1680, 'aes' -> 0.1440}

        Because the join produces a string for each possible combination, the
        size of the resulting set is equal to the product of the sizes of the
        input sets.

        To avoid blowing up the memory, the <cut_off> parameter can be used
        to filter away any resulting string that has a lesser than <cut_off>
        value.

        >>> join_fuzzy_sets([A, B, C], cut_off=0.2, j='')
        {'yjs' -> 0.2520, 'yes' -> 0.2160}

    '''

    assert 0 <= cut_off <= 1.0

    all_possibilities = itertools.product(*(fs.items() for fs in iterable))

    def xxx(elems):
        byte_seq, probs = zip(*elems)

        seq_prob = 1
        for p in probs:
            seq_prob *= p

        return byte_seq, seq_prob

    # calculate the probability for each sequence
    tmp = (xxx(elems) for elems in all_possibilities)

    # filter out the less likely and join the remain byte sequences
    tmp = ((j.join(byte_seq), pr) for byte_seq, pr in tmp if pr >= cut_off)

    # build up the fuzzy set
    return FuzzySet(dict(tmp))




import base64
import collections
import itertools
import math
import numbers

from cryptonita.stats import SequenceStatsMixin

'''
>>> # Convenient definitions
>>> from cryptonita.bytestring import as_bytes, B, ByteString, InfiniteStream

'''

def are_same_length_or_fail(a, b):
    if len(a) != len(b):
        raise ValueError("Mismatch lengths. Left string has %i bytes but right string has %i." % \
                            (len(a), len(b)))

def are_bytes_or_fail(val, name):
    if not isinstance(val, collections.ByteString):
        raise TypeError("The parameter '%s' should be an instance of %s but it is %s." % \
                            (name, bytes, type(val)))

def as_bytes(raw, encoding='ascii', return_bytestring=True):
    r'''
        Create a stream of bytes from a sequence from bytes or text (str).

         - from a <raw> sequence of bytes:

            >>> as_bytes(b'\x00\x01')
            '\x00\x01'

         - from an iterable of integers

            >>> as_bytes([0, 1])
            '\x00\x01'

         - from a ByteString

            >>> as_bytes(as_bytes(b'\x00\x01'))
            '\x00\x01'

         - from a text (str) we need to pass which encoding to use to decode
         the string to bytes:

            >>> as_bytes(u'AB', encoding='utf-8')
            'AB'

         - it is also supported an overloaded version of <encoding> to map
         a base 16, 64, ... string or bytes into a stream.

            >>> as_bytes(b'020b', encoding=16)
            '\x02\x0b'

         - we can do the same with a text (str). In this case we assume that
         the encoding to map the unicode to the raw bytes is 'ascii',
         then we use <encoding> to map it to its final state

            >>> as_bytes(u'020b', encoding=16)
            '\x02\x0b'

         - from an integer. We assume that the integer is a byte sequence of
         just one byte.

            >>> as_bytes(12)
            '\x0c'

        By default, the object returned is an enhanced version of a sequence
        of bytes. If you want the raw bytes set <return_bytestring> to False:

            >>> s = as_bytes(12)
            >>> isinstance(s, bytes) and isinstance(s, ByteString)
            True

            >>> b = as_bytes(12, return_bytestring=False)
            >>> isinstance(b, bytes) and not isinstance(b, ByteString)
            True

        '''
    # see a single byte as a byte string
    #   as_bytes(7) -> b'\x07'
    if isinstance(raw, int):
        raw = bytes([raw])

    # for a unicode, encode it to bytes
    #   as_bytes(u'text', encoding='utf8') -> b'text' (encode: utf8)
    elif isinstance(raw, str):
        if isinstance(encoding, int):
            # if the encoding is an integer means that it is
            # for decoding the string later.
            # Assume that encoding is then 'ascii'
            enc = 'ascii'
        else:
            enc = encoding

        raw = raw.encode(enc, errors='strict')

    #   as_bytes([b'\x0A', b'\x0B']) -> b'\x0a\x0b'
    #   as_bytes(b'\x00') -> b'\x00'
    else:
        raw = bytes(raw)

    # overloaded meaning of encoding, the new raw bytes are encoded
    # using base 16 (64, other) and we want to decode them.
    if isinstance(encoding, int):
        if encoding == 16:
            raw = raw.upper()
        raw = getattr(base64, 'b%idecode' % encoding)(raw)

    return ByteString(raw) if return_bytestring else raw

def load_bytes(fp, mode='rb', **k):
    ''' Open a file <fp> with mode <mode> (read - binary by default)
        and read and load each line as a ByteString.

        How each line should be processed can be controlled by the
        same parameters that can be used with as_bytes.
        The keyword parameters of load_bytes are passed to
        as_bytes directly.

        If <fp> is not a string, it is assumed that it is a file
        already open (and <mode> is ignored).
        '''
    if isinstance(fp, str):
        fp = open(fp, mode)

    return (as_bytes(line.strip(), **k) for line in fp)

# alias
B = as_bytes


class ByteString(bytes, SequenceStatsMixin):
    r'''
        Enhance a string of bytes extending it with some convenient
        methods.

        First, it is a sequence of bytes:

            >>> s = ByteString(b'\x01\x02')

            >>> isinstance(s, bytes)
            True

        It supports the same set of methods than any other bytes sequence
        does.

        Like indexing and slicing:

            >>> s[0]
            1

            >>> s[:1]
            '\x01'

            >>> len(s)
            2

        Decoding (to decode the sequence of bytes into a unicode text):

            >>> t = s.decode(encoding='ascii')
            >>> isinstance(t, str)
            True

        But it also supports encoding (to encode the sequence using a
        base 16, 64, ... encoding schema):

            >>> s.encode(base=16)
            '0102'

        See more information in the documentation of each method.

        '''

    def ngrams(self, n):
        return Ngrams(self, n)

    def nblocks(self, n):
        return Nblocks(self, n)

    # TODO cache me
    def count_1s(self):
        return sum(_number_of_1s_in_byte[b] for b in self)

    def join(self, *others):
        return B(bytes.join(self, *others))

    def __xor__(self, other):
        r'''
            Compute the xor between two strings of bytes with the same length.

                >>> a = B('1c0111001f010100061a024b53535009181c', encoding=16)
                >>> b = B('686974207468652062756c6c277320657965', encoding=16)

                >>> c = a ^ b
                >>> c.encode(16)
                '746865206B696420646F6E277420706C6179'

            If the two strings have different lengths, the operation is undefined.
            Should we repeat the shorter string to match the length of the other
            or should we truncate the longer instead to match the length of the
            shorter?

            Different lengths are not allowed

                >>> short = B('\x11\x10\x00\x00')
                >>> a ^ short
                Traceback (most recent call last):
                <...>
                ValueError: Mismatch lengths. Left string has 18 bytes but right string has 4.

            There is an exception to this rule: if one of the stream is an
            infinite sequence, then (obviously) the only logical action is to
            truncate it to match the length of the shorter sequence and compute
            the xor:

                >>> plaintext = B("Burning 'em, if you ain't quick and nimble\nI go crazy when I hear a cymbal")
                >>> key = B("ICE")

                >>> (plaintext ^ key.inf()).encode(16)
                ('0B3637272A2B2E63622C2E69692A23693A2A3C6324202D623D63343C2A26226324272765272A'
                 '282B2F20430A652E2C652A3124333A653E2B2027630C692B20283165286326302E27282F')

        '''
        if not isinstance(other, InfiniteStream):
            are_same_length_or_fail(self, other)

        return ByteString((a ^ b for a, b in zip(self, other)))

    def __rxor__(self, other):
        return self ^ other

    def __add__(self, other):
        return as_bytes(bytes.__add__(self, other))

    def __radd__(self, other):
        return as_bytes(other.__add__(self))

    def __lshift__(self, other):
        r'''
                >>> s = B("ABCD")

                >>> s << B('E')
                'BCDE'

                >>> s << B('EFG')
                'DEFG'

                >>> s << B('EFGHIJK')
                'HIJK'

        '''
        n = len(other)

        if n == 0:
            return B(self)

        if n > len(self):
            return B(other[-len(self):])

        return self[n:] + other

    def inf(self):
        return InfiniteStream(self)

    def encode(self, base):
        r'''
            Encode the byte string using a x base (base 16, base 64, ...)

                >>> B('\x01\x02').encode(16)
                '0102'

            Combine this with the decoding capabilities of B (as_bytes) to create
            a (little inefficient) conversor between bases:

                >>> b16 = b"49276d206b696c6c696e6720796f757220627" + \
                ...       b"261696e206c696b65206120706f69736f6e6f" + \
                ...       b"7573206d757368726f6f6d"
                >>> B(b16, encoding=16).encode(64)
                'SSdtIGtpbGxpbmcgeW91ciBicmFpbiBsaWtlIGEgcG9pc29ub3VzIG11c2hyb29t'

            '''

        return getattr(base64, 'b%iencode' % base)(self)

    def pad(self, n, scheme):
        r'''Pad the byte string and return a new byte string object.

            How many bytes to pad is controlled by <n>, how we should
            do it is controlled by <scheme>

            For example:

                >>> B('AAAAAAAAAAAA').pad(16, 'pkcs#7')
                'AAAAAAAAAAAA\x04\x04\x04\x04'

        '''
        if scheme == 'pkcs#7':
            npad = n - (len(self) % n)
            if npad == 0:
                npad = n

            padding = bytes([npad]) * npad
        else:
            raise ValueError("Unknow padding scheme '%s'" % scheme)

        return as_bytes(self + padding)

    def unpad(self, scheme):
        r'''
                >>> padded = B('AAAAAAAAAAAA').pad(16, 'pkcs#7')
                >>> padded.unpad('pkcs#7')
                'AAAAAAAAAAAA'

        '''
        if scheme == 'pkcs#7':
            n = self[-1]
            assert n <= 64

            if self[-n:] != B(n) * n:
                raise ValueError("Bad padding '%s' with last byte %x" %
                                        (scheme, n))

            return self[:-n]
        else:
            raise ValueError("Unknow padding scheme '%s'" % scheme)



class Ngrams(SequenceStatsMixin):
    ''' N-grams view of a byte string.

        A n-gram is a sequence of n bytes that overlaps
        each other.

            >>> s = B('ABCDE')
            >>> ngrams = s.ngrams(3)

            >>> list(ngrams)
            ['ABC', 'BCD', 'CDE']

        Shorter and larger sequences are ok

            >>> list(B('ABC').ngrams(3))
            ['ABC']

            >>> list(B('ABCDEFGHI').ngrams(3))
            ['ABC', 'BCD', 'CDE', 'DEF', 'EFG', 'FGH', 'GHI']

        But too short aren't

            >>> list(B('AB').ngrams(3))
            Traceback <...>
            <...>
            ValueError: The byte string has only 2 bytes. It is not possible to create even a single ngram of length 3.

        A n-gram sequence supports:

            >>> len(ngrams)
            3

            >>> ngrams[1]
            'BCD'

        But slicing is not supported (yet)

            >>> ngrams[1:2]
            Traceback <...>
            <...>
            TypeError: We don't support slicing (yet).

        This also support some basic statistics:

            >>> ngrams = B('ABABCABABC').ngrams(2)
            >>> ngrams.freq()
            Counter({b'AB': 4, b'BA': 2, b'BC': 2, b'CA': 1})

            >>> ngrams.most_common(n=2)
            ('AB', 'BA')

    '''
    __slots__ = ('_raw', '_n')

    def __init__(self, raw, n):
        if len(raw) < n:
            raise ValueError("The byte string has only %i bytes. It is not possible to create even a single ngram of length %i." % (len(raw), n))

        self._raw = raw
        self._n = n

    def __iter__(self):
        s = self._raw
        n = self._n
        return (s[i:i+n] for i in range(len(self)))

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            raise TypeError("We don't support slicing (yet).")

        return self._raw[idx:idx+self._n]

    def __len__(self):
        return len(self._raw) - self._n + 1

class Nblocks(SequenceStatsMixin):
    ''' N-blocks view of a byte string.

        A n-block is a sequence of n bytes that do not overlap
        each other.

            >>> s = B('ABCDEF')
            >>> nblocks = s.nblocks(3)

            >>> list(nblocks)
            ['ABC', 'DEF']

        When the length of a sequence is not a multiple of n, the last
        block will have less than n bytes

            >>> list(B('ABC').nblocks(3))
            ['ABC']

            >>> list(B('ABCDEFGHIJK').nblocks(3))
            ['ABC', 'DEF', 'GHI', 'JK']

            >>> len(B('ABCDEFGHIJK').nblocks(3))
            4


        A n-blok sequence supports:

            >>> len(nblocks)
            2

            >>> nblocks[1]
            'DEF'

        But slicing is not supported (yet)

            >>> nblocks[1:2]
            Traceback <...>
            <...>
            TypeError: We don't support slicing (yet).


        This also support some basic statistics:

            >>> nblocks = B('ABABCABAABBA').nblocks(2)
            >>> nblocks.freq()
            Counter({b'AB': 3, b'BA': 2, b'CA': 1})

            >>> nblocks.most_common(n=2)
            ('AB', 'BA')

    '''
    def __init__(self, raw, n):
        self._raw = raw
        self._n = n

    def __iter__(self):
        s = self._raw
        n = self._n
        return (s[i*n:(i+1)*n] for i in range(len(self)))

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            raise TypeError("We don't support slicing (yet).")

        n = self._n
        return self._raw[idx*n:(idx+1)*n]

    def __len__(self):
        return math.ceil(len(self._raw) / self._n)

class InfiniteStream(bytes):
    def __getitem__(self, idx):
        if isinstance(idx, slice):
            raise TypeError("We don't support slicing (yet).")

        return bytes.__getitem__(self, idx % bytes.__len__(self))

    def __iter__(self):
        r'''
            Return an iterator of this infinite stream of bytes.
            Basically we take a finite sequence and we repeat it
            infinite times:

                >>> s = InfiniteStream(b'\x01\x02')

                >>> i = iter(s)
                >>> next(i), next(i), next(i), next(i)
                (1, 2, 1, 2)

            '''
        return itertools.cycle(bytes.__iter__(self))

    def __repr__(self):
        return '.. ' + bytes.__repr__(self) + ' ..'

    def __len__(self):
        return math.inf

_number_of_1s_in_byte = [0] * 256
for i in range(len(_number_of_1s_in_byte)):
    _number_of_1s_in_byte[i] = (i & 1) + _number_of_1s_in_byte[i >> 1]




import json
import math
import itertools
from collections import Counter

import scipy.stats as stats
from langdetect import detect_langs

from cryptonita.bytestring import B, are_bytes_or_fail

'''
>>> # Convenient definitions
>>> from cryptonita.bytestring import B
>>> from cryptonita.scoring import *

'''


def hamming_distance(m1, m2):
    r'''
        Return the Hamming distance between <m1> and <m2>.

        >>> m1 = B('this is a test')
        >>> m2 = B('wokka wokka!!!')

        >>> hamming_distance(m1, m2)
        37

        The Hamming (or edit) distance is the count of how many bits
        these two string differ. It is defined for strings of the same
        lengths, so the following will fail:

        >>> hamming_distance(m1, m1 + m1)
        Traceback (most recent call last):
        <...>
        ValueError: Mismatch lengths. Left string has 14 bytes but right string has 28.

    '''

    x = m1 ^ m2
    return x.count_1s()

def all_ascii_printable(m):
    ''' Score with 1 if the message has only ASCII printable characters.

            >>> import string

            >>> message = B(string.printable, encoding='ascii')
            >>> all_ascii_printable(message)
            1

        Score with 0 otherwise

            >>> message = B(string.printable + "\x04", encoding='ascii')
            >>> all_ascii_printable(message)
            0

    '''
    are_bytes_or_fail(m, 'm')
    return 1 if all((32 <= b <= 126 or 9 <= b <= 13) for b in m) else 0

def is_language(m, language):
    return detect_langs(m)[language]

def ngram_entropy_score(m, N=1):
    r'''
        Take the frequency of each ngram of <N> bytes of length from
        <m>, and estimate the entropy.

        >>> truly_random = B('3nQucdIjAbP1n8wZ7c+xs4KoRPo6U1GP3jJllAzU24rYiCo6' +\
        ...                  'ZOgWt/zq5nG0XyNKPO95dBytg90rXxGDJGzRCNhiFX9GLZgV' +\
        ...                  'NKxbbbvrZ3bBc1b6TcthBvMWYvGAkdmOeR/TDaJgd0/x5nt5' +\
        ...                  'SiZ7Aq1+n2Eood2LxCLYCiytrp8=', encoding=64)

        >>> ngram_entropy_score(truly_random, N=1)
        4.55<...>

        Lower values are associated with string less random.

        >>> no_random = B([0b001, 0b010, 0b110, 0b001] * 64)
        >>> ngram_entropy_score(no_random, N=1)
        1.039<...>

        It is quite useful to distinguish short random sequences from not-much
        random sequences.

        It is related with index_of_coincidence, but in the practice, entropy
        is slower and it is less discriminant.

        Nota bene: the entropy is calculated as:

            S = -sum(pk * log(pk))

        where pk is the frequency of the kth ngram and log is the natural
        logarithm (base e).

    '''
    freq = list(m.ngrams(N).freq().values())
    return stats.entropy(freq)

def yes_no_score(m, yes_prob=0.5):
    r'''
        Count how many 1s (yes or successes) the <m> strings has (a byte
        string that it will be seing like a string of bits).

        Then, under the hypothesis of that the sequence <m> is a Bernulli
        process (a random uniform string), <yes_prob> is the hypothetical
        probability of yes or success.

        >>> truly_random = B('3nQucdIjAbP1n8wZ7c+xs4KoRPo6U1GP3jJllAzU24rYiCo6' +\
        ...                  'ZOgWt/zq5nG0XyNKPO95dBytg90rXxGDJGzRCNhiFX9GLZgV' +\
        ...                  'NKxbbbvrZ3bBc1b6TcthBvMWYvGAkdmOeR/TDaJgd0/x5nt5' +\
        ...                  'SiZ7Aq1+n2Eood2LxCLYCiytrp8=', encoding=64)

        Under this hypothetical (named the null hypothesis), return the
        probability of *reject* the hypothesis.

        >>> yes_no_score(truly_random, yes_prob=0.5)
        0.488<...>

        The value returned is the probability of reject the hypothesis or
        the p-value [1].

        Even if the p-value is really small so we cannot reject the hypothesis,
        we *cannot* say then that <m> *is* truly uniformely random;
        We *only* can say that we don't have evidence to *reject*
        the hypothesis.

        Only high values allow us to reject the hypothesis, nothing else.

        >>> no_random = B([0b001, 0b010, 0b110, 0b001] * 64)
        >>> yes_no_score(no_random, yes_prob=0.5)
        1.0

        [The p-value] "does not tell us what we want to know,
        and we so much want to know what we want to know that,
        out of desperation, we nevertheless believe that it does"
                                                        (Cohen, 1994).

        References:

        [1] p-value: https://en.wikipedia.org/wiki/P-value
        [2] binomial test: https://en.wikipedia.org/wiki/Binomial_test
    '''

    sucessess = m.count_1s()
    bits = len(m) * 8

    return 1 - stats.binom_test(sucessess, n=bits)

def index_of_coincidence(m, N=1, mode='ngrams'):
    r'''
        Take the string of bytes <m> and see it as a sequence of ngrams
        of length <N> if <mode> is 'ngrams' or see it as a sequence of
        non-overlaping blocks of length <N> if <mode> is 'nblocks'.

        Then, see how many ngrams are repeated: count the coincidences; higher
        values means that the <m> sequence is less random. [1]

        For example, in the given file are 327 random strings. One of them
        is actually a message encrypted by doing a xor with a key of just 1 byte.
        Therefore, it should have more coincidences than the rest of the strings.

        >>> ciphertexts = open('data/4.txt', 'rb').read().strip().split(b'\n')
        >>> ciphertexts = [B(c, encoding=16) for c in ciphertexts]

        >>> scores_and_indexes = [(index_of_coincidence(c), i) \
        ...                         for i, c in enumerate(ciphertexts)]

        Remember, higher values are better

        >>> max(scores_and_indexes)
        (0.045977<...>, 170)

        Let's check if the 170th string is our ciphered message

        >>> key = B('5') # i am doing a little cheat here to speed up the things

        >>> ciphertexts[170] ^ key.inf()
        'Now that the party is jumping\n'

        References:

        [1] Index of coincidence: https://en.wikipedia.org/wiki/Index_of_coincidence

    '''
    if mode == 'ngrams':
        seq = m.ngrams(N)
    elif mode == 'nblocks':
        seq = m.nblocks(N)
    else:
        raise ValueError("Unknow mode %s" % mode)

    freqs = seq.freq().values()
    text_len = len(m) - N + 1
    # TODO talk about why the c factor should not be used:
    # - it makes the final number a non-prob value (greather than 1)
    # - it is just a scalar factor to separate one value from the other but
    #   this is a human invention, the machine has no problem to see the
    #   difference of 0.0001 and 0.000011.
    return sum(f * (f-1) for f in freqs) / (text_len * (text_len - 1))


def count_duplicated_blocks(s, block_size, distance=None, indexes=False):
    r'''Count how many blocks are repeated or duplicated.

            >>> s = B('AABBAACCCCDDAADDAA')
            >>> count_duplicated_blocks(s, block_size=2)
            5

        If the <distance> parameter is given, a block is considered
        duplicated if another block is the same *and* it is <distance>
        blocks of distance.

        <distance> equal to 0 means "two consecutive blocks"

            >>> # the CCs
            >>> count_duplicated_blocks(s, block_size=2, distance=0)
            1

            >>> # the first 2 and last 2 AAs and the DDs
            >>> count_duplicated_blocks(s, block_size=2, distance=1)
            3

            >>> # the first 2 AAs with the last 2 AAs
            >>> count_duplicated_blocks(s, block_size=2, distance=5)
            2

    '''
    # TODO i'm not resistent to possible false positive!
    blocks = list(s.nblocks(block_size))

    if distance is not None:
        # consecutive blocks if distance is 0
        # 1 block between if distance is 1, ...
        pairs = zip(blocks[:-1-distance], blocks[1+distance:])

        count = 0
        for a, b in pairs:
            if a == b:
                count += 1

        return count

    else:
        # count any duplicated block, no matter how far is it from
        # its copy
        before = len(blocks)
        after  = len(set(blocks))

        return before - after

def key_length_by_hamming_distance(length, ciphertext):
    ''' Score how much likely is <length> to be the length of the key
        which was used to encrypt and obtain the ciphertext.

        This estimate is done using the hamming distance.

        If K is a random uniform distributed key space, k one key from that
        space and ki and kj two bytes of that key.

        For the ci and cj ciphertext bytes we have:
            pi = ci ^ ki   and   pj = cj ^ kj

        then, the hamming distance will be:

            h(ci, cj) = sum_1s((pi ^ ki) ^ (pj ^ kj))
                      = sum_1s((pi ^ pj) ^ (ki ^ kj))

        now if ki == kj, the hamming distance will be

            h(ci, cj) = sum_1s((pi ^ pj))

        If the plaintext space is *not* uniform distributed, the count of 1s
        of pi ^ pj will be smaller (pi and pj are similar)

        If ki != kj, then ki ^ kj is a random uniform distributed byte so
          (pi ^ pj) ^ (ki ^ kj) is also a random uniform distributed byte
        and the expected count of 1s is half (4 bits for a 8-bits byte)

        With this we can discriminate when ki == kj or not

        However this assumes two thing:
         - K is a random uniform distributed key string
         - P is *not* a random uniform distributed plaintext string.

        If K is not a random uniform distributed key space, the hamming
        distance may not be the best discriminant.
        '''

    l = length
    if len(ciphertext) < l * 2:
        raise ValueError("The ciphertext is too short to guess the key's length and it is impossible to see if a key of %i bytes could be possible." % l)

    distance = hamming_distance(B(ciphertext[:l]), B(ciphertext[l:l*2]))
    return 1 - (distance / (l*8))


def key_length_by_ic(length, ciphertext, N=1):
    ''' Score the possible <length> of the key that was used to encrypt
        and obtain the <ciphertext> using the Index of Coincidence (IC).

        Pick bytes from <ciphertext> that are <length> bytes of distance:
            1, l, 2l, ...

        and compute the IC.
        '''
    return index_of_coincidence(B(ciphertext[::length]), N)


# (b - a + 1)**2 - 1 / 12  = (b==1, a==0) -> 3/12
# Chebyshev's_inequality

# P [ |X - u| >= kd ] <= 1/k**2    (u = mean, d = sqrt(variance), k >= 1, X random variable)
# P [ |X - u| >= a ]  <= v / a**2  (v = var, a >= 1)



import collections

class SequenceStatsMixin:
    def freq(self):
        return collections.Counter(self)

    def most_common(self, n):
        elems, _ = zip(*self.freq().most_common(n))
        return elems




import cryptonita.scoring as scoring
from cryptonita.fuzzy_set import FuzzySet
from cryptonita.bytestring import B, are_bytes_or_fail

from itertools import product

'''
>>> # Convenient definitions
>>> from cryptonita.bytestring import B
>>> from cryptonita.attacks import brute_force, freq_attack  # byexample: +timeout=10

'''

def brute_force(ciphertext, score_func, key_space=1,
                        min_score=0.25):
    r'''Guess what key was used to xor the <ciphertext>.
        Guessing means try every single possible key so we need to score
        each try with <score_func> to see what key is really useful.

        For example, if we know that the plaintext is a Bitmap image
        we can score a decrypted message like:

            >>> def is_bmp(m):
            ...     return 1 if m[:2] == b"BM" else 0

        because all the Bitmap images starts with the magic bytes 'BM'.

        Then, we guess the key:

            >>> ciphertext = B('\x1a\x15XXXY')
            >>> guessed_keys = brute_force(ciphertext, is_bmp)

            >>> guessed_keys
            {'X' -> 1.0000}


        Not always the guess is perfect. Because of this we return a FuzzySet
        object that each guessed key with its likehood.

        If <key_space> is an integer, the key space is all possible
        keys of <key_space> bytes.
        It is advised to keep <key_space> less than 3... may be 4.
        Also, use <min_score> to drop all the keys with score less than
        <min_score> to keep the memory usage at minumum.

            >>> ciphertext = B('\x1a\x14XYXX')
            >>> brute_force(ciphertext, is_bmp, key_space=2)
            {'XY' -> 1.0000}

        Otherwise, <key_space> must be an iterable of bytes: each string of bytes
        is a key to try (we will try all of them in the given order)

            >>> ciphertext = B('\x1a\x15XXXY')
            >>> possible_keys = (B(k) for k in ('Q', 'X', 'Z'))
            >>> brute_force(ciphertext, is_bmp, key_space=possible_keys)
            {'X' -> 1.0000}

    '''
    assert 0.0 <= min_score <= 1.0
    are_bytes_or_fail(ciphertext, 'ciphertext')

    prob = {}
    if isinstance(key_space, int):
        key_space = (B(k) for k in product(range(256), repeat=key_space))

    elif isinstance(key_space, FuzzySet):
        prob = key_space

    keys = FuzzySet(((k, score_func(ciphertext ^ k.inf()) * prob.get(k, 1)) for k in key_space),
                        pr='tuple',
                        min_membership=min_score)
    return keys

def freq_attack(ciphertext, most_common_plain_ngrams, cipher_ngram_top=1):

    r'''Try to break the ciphering doing a frequency attack.
        The idea is that the plain text has some ngrams more frequent than
        others and that is reflected in the <ciphertext>.

        Then, we can assume that at least one of the <cipher_ngram_top> most
        common cipher ngrams is one of the <most_common_plain_ngrams> encrypted
        with the secret key.

        >>> most_common_plain_ngrams = [B(b) for b in b'etaoin shrdlu']
        >>> cipher_ngram_top = 1

        So, for all the most common plain ngrams p and cipher ngrams c we can
        propose (c ^ p) = k.

        >>> ciphertext = B('1b37373331363f78151b7f2b783431333d78397828372d363c78373e783a393b3736',
        ...                     encoding=16)

        >>> proposed_keys = freq_attack(ciphertext, most_common_plain_ngrams)

        >>> proposed_keys       # byexample: +norm-ws
        {<...>'X' -> 1.000<...>}


        Then we can use them to guess the secret key

        >>> from cryptonita.scoring import all_ascii_printable
        >>> key = brute_force(ciphertext, key_space=proposed_keys.keys(),
        ...                 score_func=all_ascii_printable).most_likely()

        And decrypt the message

        >>> ciphertext ^ key.inf()
        "Cooking MC's like a pound of bacon"

        Nice reading: http://norvig.com/mayzner.html
    '''

    # aliases
    T = cipher_ngram_top

    are_bytes_or_fail(ciphertext, 'ciphertext')

    keys = FuzzySet()

    prob = {}
    if isinstance(most_common_plain_ngrams, FuzzySet):
        prob = most_common_plain_ngrams

    # XXX despite the code supports different values of N in the same
    # run, that is, we support do an freq attack using ngrams of different
    # lengths, it is not clear how we should merge the results.
    #
    # Using ngrams of length 1, we can propose keys of 1 byte. If we use
    # ngrams of length 2, we can propose keys of 2 bytes.
    #
    # Now, if we have propose keys of 1 and 2 bytes how we merge them?
    # If the real secret key is of 1 byte, we should discard all the proposed
    # keys of 2 bytes that don't have those 2 bytes equals.
    #
    # Then, we will have a proposed key X (1 byte) and a proposed key XX (2 bytes)
    # but they are the same. So should we remove them too?
    # But if we do, we removed all the proposed keys of 2 bytes, what was the
    # point? It will only be useful if we don't find X but XX or we don't find XX
    # and we do find X.
    for N in (1,):
        # count all the possible ngrams of N bytes of length of
        # the ciphertext
        # then, pick the T most common and therefore the most likely to be
        # the encrypted version of the most common plain ngrams
        _cipher_ngrams = ciphertext.ngrams(N).most_common(T)

        # most common plain ngrams of N bytes of length
        _plain_ngrams = filter(lambda ngram: len(ngram) == N,
                                            most_common_plain_ngrams)

        # if our hypthesis is correct, at least one of the c cipher ngrams
        # will be (p ^ k) where p is one of the p plain ngrams
        # if this is true, one of the 'proposed keys' keys will be the real
        # secret key k
        #
        tmp = FuzzySet(((c ^ p, prob.get(p, 1)) for c, p in product(_cipher_ngrams,
                                                                    _plain_ngrams)),
                                pr='tuple')
        keys.update(tmp)

    return keys

def decrypt_ecb_tail(alignment, block_size, encryption_oracle, limit=None):
    align_test_block = B("A" * alignment)

    test_block = B("A" * (block_size - 1))
    align_target_block = B(test_block) # copy
    distance = 0

    decrypted_bytes = []
    i = 0
    eof = False
    while not eof and (i < limit if limit else True):
        i += 1
        eof = True
        for b in range(256):
            b = B(b)

            # propose the following choosen plaintext:
            #
            #   |-------|-------|-------|------
            #    ....AAA AAAAAAb AAAAAA? .....
            #       |       |       |
            #       |       |  a block identical to the test block except
            #       |       |  the last byte that's unknow to us (to be decypted)
            #       |  a "test" block: a full block where the last byte
            #       |  is our guessed byte (if 'b' is equal to '?' our guess is correct)
            #    padding block for alignment purposes
            tmp = align_test_block + test_block + b + align_target_block
            c = encryption_oracle(tmp)

            if scoring.count_duplicated_blocks(c, block_size, distance):
                # two block had collided to a <distance> blocks of distance
                # that means that out guess 'b' matched with the unknow
                # bytes '?' effectively decrypting it
                eof = False
                decrypted_bytes.append(b)

                # the test block shift to the left one byte, now that we
                # now that 'b' is the correct byte we are reserving one byte
                # on the right for the next guess
                #           |-------|
                #            AAAAAAb  -> b is guessed ok (G)
                #           |-------|
                #            AAAAAAG  -> shift the block to make room
                #           |-------|
                #            AAAAAG   -> the byte missing will be filled
                #                        with the next guess 'b'
                test_block = test_block << b

                if len(align_target_block) == 0:
                    align_target_block = B("A" * (block_size - 1))
                    distance += 1
                else:
                    align_target_block = align_target_block[:-1]

                break

    return B(b''.join(decrypted_bytes))

def guess_key_length(ciphertext, length_space, score_func, min_score=0.5, **score_func_params):
    assert 0.0 <= min_score <= 1.0
    are_bytes_or_fail(ciphertext, 'ciphertext')

    if isinstance(length_space, int):
        length_space = range(1, length_space+1)

    params = score_func_params
    lengths = FuzzySet(((l, score_func(l, ciphertext, **params)) for l in length_space),
                        pr='tuple',
                        min_membership=min_score)
    return lengths




# https://packaging.python.org/en/latest/distributing.html
# https://github.com/pypa/sampleproject

from setuptools import setup, find_packages
from codecs import open
from os import path, system

import sys, re

here = path.abspath(path.dirname(__file__))

# load __version__, __doc__, _author, _license and _url
exec(open(path.join(here, 'cryptonita', '__init__.py')).read())

long_description = __doc__

install_deps=[
        'scipy',
        'langdetect',
        ]

setup(
    name='cryptonita',
    version=__version__,

    description=__doc__,
    long_description=long_description,

    #url=_url,

    # Author details
    author=_author,
    author_email='use-github-issues@example.com',

    license=_license,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',

        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',

        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        'Programming Language :: Python :: 3',
    ],

    python_requires='>=3.3',
    install_requires=install_deps,

    keywords='crypto cryptography crypto-analysis matasano',

    packages=['cryptonita'],
)

